#include "PCH.h"

#include "SystemImpl.h"

#include "GraphicMgrImpl.h"
#include "AudioMgrImpl.h"
#include "MusicMgrImpl.h"
#include "InputMgrImpl.h"
#include "TaskMgrImpl.h"
#include <stdexcept>

#if defined(WIN32) || defined(_WIN32)
#	include <windows.h>
#endif

UQ::System* UQ::System::pSystem_ = nullptr;
CIniFileA UQ::iniFile("system.ini");
//========================================================================
void UQ::showError(const std::exception& ex)
{
#if defined(WIN32) || defined(_WIN32)
	MessageBoxA(0, ex.what(), "UQGL", MB_OK | MB_ICONEXCLAMATION);
#elif
	std::fprintf(stderr, "%s\n", ex.what());
#endif
}

//========================================================================
UQ::System* UQ::System::getInstance(void)
{
	if(!pSystem_) pSystem_ = new SystemImpl();
	return pSystem_;
}
//========================================================================
void UQ::System::deleteInstance(void)
{
	delete pSystem_;
	pSystem_ = nullptr;
}

//========================================================================
UQ::SystemImpl::SystemImpl() :
	pGraphicMgr_(nullptr),
	pAudioMgr_(nullptr),
	pInputMgr_(nullptr),
	pTaskMgr_(nullptr),
	isInitialized_(false),
	frameCount_(0)
{
}

//========================================================================
UQ::SystemImpl::~SystemImpl()
{
}

//========================================================================
void UQ::SystemImpl::initialize(int windowWidth, int windowHeight, const char* windowTitle, const char* systemFontFileName, int systemFontWidth, int systemFontHeight)
{
	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK) < 0){
		throw std::runtime_error("SDL ̏Ɏs܂B");
	}
	SDL_WM_SetCaption(windowTitle, nullptr);
	SDL_ShowCursor(SDL_DISABLE);

	pGraphicMgr_ = std::unique_ptr<GraphicMgrImpl>(new GraphicMgrImpl(windowWidth, windowHeight, systemFontFileName, systemFontWidth, systemFontHeight));
	pAudioMgr_ = std::unique_ptr<AudioMgrImpl>(new AudioMgrImpl());
	pInputMgr_ = std::unique_ptr<InputMgrImpl>(new InputMgrImpl());
	pTaskMgr_ = std::unique_ptr<TaskMgrImpl>(new TaskMgrImpl());
	isInitialized_ = true;
}

//========================================================================
void UQ::SystemImpl::run(void)
{
	ASSERT_IS_INTIALIZED;

	bool isConfiguringVolume = false, isConfiguringKey = false;

	while(doEvents()){
		Uint32 startTime = SDL_GetTicks();

		pInputMgr_->update();
		if(pInputMgr_->isPressed(InputMgr::K_F11)) pGraphicMgr_->changeDisplayMode();
		pGraphicMgr_->begin();
		if(isConfiguringVolume){
			isConfiguringVolume = configureVolume();
		}else if(isConfiguringKey){
			isConfiguringKey = configureKey();
		}else{
			if(pInputMgr_->isPressed(InputMgr::K_ESC)) quit();
			if(pInputMgr_->isPressed(InputMgr::K_F1)) isConfiguringKey = true;
			if(pInputMgr_->isPressed(InputMgr::K_F2)) isConfiguringVolume = true;
			pTaskMgr_->executeAll();
			pTaskMgr_->optimize();
		}
		pGraphicMgr_->drawWindowBox();
		pGraphicMgr_->end();
		
		frameCount_++;

		SDL_Delay(1);
		Uint32 endTime = SDL_GetTicks();
		wait(1000 / 60 - (endTime - startTime));
	}
}

//========================================================================
bool UQ::SystemImpl::doEvents(void)
{
	ASSERT_IS_INTIALIZED;

	SDL_Event event;
	while(SDL_PollEvent(&event)){
		if(event.type == SDL_QUIT) return false;
	}
	return true;
}

//========================================================================
void UQ::SystemImpl::wait(int milliseconds)
{
	ASSERT_IS_INTIALIZED;

	if(milliseconds <= 0) return;
	Uint32 startTime = SDL_GetTicks();
	do{
		SDL_Delay(1);
	}while(SDL_GetTicks() < milliseconds + startTime);
}

//========================================================================
UQ::GraphicMgr* UQ::SystemImpl::getGraphicMgr(void) const{ASSERT_IS_INTIALIZED; return pGraphicMgr_.get();}
UQ::AudioMgr*   UQ::SystemImpl::getAudioMgr(void)   const{ASSERT_IS_INTIALIZED; return pAudioMgr_.get();}
UQ::InputMgr*   UQ::SystemImpl::getInputMgr(void)   const{ASSERT_IS_INTIALIZED; return pInputMgr_.get();}
UQ::TaskMgr*    UQ::SystemImpl::getTaskMgr(void)    const{ASSERT_IS_INTIALIZED; return pTaskMgr_.get();}

//========================================================================
bool UQ::SystemImpl::configureVolume(void)
{
	float volume = pAudioMgr_->getVolume();
	if(pInputMgr_->isDown(InputMgr::K_LEFT )) volume -= 0.01f;
	if(pInputMgr_->isDown(InputMgr::K_RIGHT)) volume += 0.01f;
	if(volume > 1.f) volume = 1.f;
	if(volume < 0.f) volume = 0.f;
	pAudioMgr_->setVolume(volume);

	static const int X = 32, Y = 16;
	static const int FONT_WIDTH = pGraphicMgr_->getSystemFontWidth(), FONT_HEIGHT = pGraphicMgr_->getSystemFontHeight();
	static const int BAR_Y = Y + FONT_HEIGHT * 2, BAR_WIDTH = FONT_WIDTH * 20, FRAME_WIDTH = 2;
	
	pGraphicMgr_->outputString(X, Y, "<< Volume Configuration >>");
	pGraphicMgr_->outputString(X, BAR_Y, "Volume");
	pGraphicMgr_->outputString(X + FONT_WIDTH * 8 + BAR_WIDTH, BAR_Y, makeString("%d", int(volume * 100)).c_str());
	
	pGraphicMgr_->unsetTexture();
	Polygon rect;
	// wio[
	rect.setRectangle(X + FONT_WIDTH * 7, BAR_Y, BAR_WIDTH, FONT_HEIGHT);
	rect.setColor(0xFF0000FF);
	pGraphicMgr_->drawPolygon(rect);
	// {[o[
	rect.setRectangle(X + FONT_WIDTH * 7 + FRAME_WIDTH, BAR_Y + FRAME_WIDTH, int(volume * (BAR_WIDTH - FRAME_WIDTH * 2 - 1)) + 1, FONT_HEIGHT - FRAME_WIDTH * 2);
	rect.setColor(0xFFFFFFFF);
	pGraphicMgr_->drawPolygon(rect);

	pGraphicMgr_->outputString(X, BAR_Y + FONT_HEIGHT * 2, "Adjust : LEFT( <- ) / RIGHT( -> )");
	pGraphicMgr_->outputString(X, BAR_Y + FONT_HEIGHT * 3, "Apply  : ENTER");

	if(pInputMgr_->isPressed(InputMgr::K_ENTER) || pInputMgr_->isPressed(InputMgr::K_F2)) return false;
	return true;
}
//========================================================================
bool UQ::SystemImpl::configureKey(void)
{
	struct Current{
		Current() : padNo(1), button(0), assign(0), isSelected(false){}
		int padNo;
		int button, assign;
		bool isSelected;
	};

	static Current* pCurrent = nullptr;
	if(!pCurrent) pCurrent = new Current();

	if(!pCurrent->isSelected){
		if(pInputMgr_->isPressed(InputMgr::K_LEFT)){
			if(pCurrent->button == -1){
				if(pCurrent->padNo > 1) pCurrent->padNo--;
			}else if(pCurrent->button < 4 + InputMgr::VC_BTN_COUNT_){
				if(pCurrent->assign > 0) pCurrent->assign--;
			}
		}
		if(pInputMgr_->isPressed(InputMgr::K_RIGHT)){
			if(pCurrent->button == -1){
				if(pCurrent->padNo < InputMgr::VC_COUNT_) pCurrent->padNo++;
			}else if(pCurrent->button < 4 + InputMgr::VC_BTN_COUNT_){
				if(pCurrent->assign < InputMgr::ASSIGN_COUNT_ - 1) pCurrent->assign++;
			}
		}
		if(pCurrent->button > -1){
			if(pInputMgr_->isPressed(InputMgr::K_UP)){
				while(true){
					pCurrent->button--;
					if(pCurrent->button == -1) break;
					if(std::strlen(pInputMgr_->getAssignName(static_cast<InputMgr::ButtonID>(pCurrent->button + pInputMgr_->VC_LEFT(pCurrent->padNo)))) > 0) break;
				}
			}
		}
		if(pCurrent->button < 4 + InputMgr::VC_BTN_COUNT_){
			if(pInputMgr_->isPressed(InputMgr::K_DOWN)){
				while(true){
					pCurrent->button++;
					if(pCurrent->button == 4 + InputMgr::VC_BTN_COUNT_) break;
					if(std::strlen(pInputMgr_->getAssignName(static_cast<InputMgr::ButtonID>(pCurrent->button + pInputMgr_->VC_LEFT(pCurrent->padNo)))) > 0) break;
				}
			}
		}
		if(pInputMgr_->isPressed(InputMgr::K_ENTER)) pCurrent->isSelected = true;
	}else{
		for(int k = InputMgr::VC1_LEFT - 1; k > InputMgr::ID_DISABLE; k--){
			if(pInputMgr_->isPressed(static_cast<InputMgr::ButtonID>(k))){
				pCurrent->isSelected = false;
				pInputMgr_->assign(static_cast<InputMgr::ButtonID>(pCurrent->button + pInputMgr_->VC_LEFT(pCurrent->padNo)), pCurrent->assign, ((k == InputMgr::K_ESC) ? InputMgr::ID_DISABLE : static_cast<InputMgr::ButtonID>(k)));
			}
		}
	}

	static const int X = 32, Y = 16;
	static const int FONT_WIDTH = pGraphicMgr_->getSystemFontWidth(), FONT_HEIGHT = pGraphicMgr_->getSystemFontHeight();
	static const int ASSIGN_X = X + FONT_WIDTH * 20, ASSIGN_Y = Y + FONT_HEIGHT * 2, ASSIGN_WIDTH = FONT_WIDTH * 12;
	static const int APPLY_Y = ASSIGN_Y + FONT_HEIGHT * (4 + InputMgr::VC_BTN_COUNT_ + 2);

	// J[\
	pGraphicMgr_->unsetTexture();
	Polygon rect;
	rect.setRectangle(X + FONT_WIDTH * (46 + (pCurrent->padNo - 1) * 3), Y, FONT_WIDTH * 3, FONT_HEIGHT);
	rect.setColor((pCurrent->button == -1) ? 0xFFFF0000 : 0xFF0000FF);
	pGraphicMgr_->drawPolygon(rect);
	if(pCurrent->button == -1){
		// select VC
		// do nothing.
	}else if(pCurrent->button == 4 + InputMgr::VC_BTN_COUNT_){
		// apply
		rect.setRectangle(X, APPLY_Y, FONT_WIDTH * 5, FONT_HEIGHT);
		rect.setColor(0xFF0000FF);
	}else{
		// assign
		rect.setRectangle(ASSIGN_X + pCurrent->assign * ASSIGN_WIDTH, ASSIGN_Y + pCurrent->button * FONT_HEIGHT + ((pCurrent->button < 4) ? 0 : FONT_HEIGHT), ASSIGN_WIDTH, FONT_HEIGHT);
		rect.setColor((pCurrent->isSelected) ? 0xFFFF0000 : 0xFF0000FF);
	}
	pGraphicMgr_->drawPolygon(rect);

	pGraphicMgr_->outputString(X, Y, "<< Key Configuration >>");
	pGraphicMgr_->outputString(X + FONT_WIDTH * 27, Y, "Virtual Controller  1  2");

	for(int i = 0; i < 4 + InputMgr::VC_BTN_COUNT_; i++){
		InputMgr::ButtonID id = static_cast<InputMgr::ButtonID>(i + pInputMgr_->VC_LEFT(pCurrent->padNo));
		Color color = (std::strlen(pInputMgr_->getAssignName(id)) > 0) ? 0xFFFFFFFF : 0xFF888888;
		int y = ASSIGN_Y + i * FONT_HEIGHT + ((i < 4) ? 0 : FONT_HEIGHT);
		if(i < 4){
			pGraphicMgr_->outputString(X, y, makeString("    %s", pInputMgr_->getAssignName(id)).c_str(), color);
		}else{
			pGraphicMgr_->outputString(X, y, makeString("%2d: %s", i - 3, pInputMgr_->getAssignName(id)).c_str(), color);
		}
		for(int j = 0; j < InputMgr::ASSIGN_COUNT_; j++){
			pGraphicMgr_->outputString(ASSIGN_X + ASSIGN_WIDTH * j, y, pInputMgr_->getButtonName(pInputMgr_->getAssign(id, j)), color);
		}
	}
	pGraphicMgr_->outputString(X, APPLY_Y, "Apply");
	pGraphicMgr_->outputString(X, APPLY_Y + FONT_HEIGHT * 2, "Hit ENTER to select. -> Push a key or a button to assign.");
	pGraphicMgr_->outputString(X, APPLY_Y + FONT_HEIGHT * 3, "                     -> Push ESC to cancel the assignment.");

	if(pInputMgr_->isPressed(InputMgr::K_F1)
		|| (pInputMgr_->isPressed(InputMgr::K_ENTER) && pCurrent->button == 4 + InputMgr::VC_BTN_COUNT_))
	{
		delete pCurrent;
		pCurrent = nullptr;
		return false;
	}
	return true;
}
